'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var events = require('events');
var createDebug = _interopDefault(require('debug'));
var Promise = _interopDefault(require('bluebird'));
var got = _interopDefault(require('got'));
var login = _interopDefault(require('plug-login'));
var socket = _interopDefault(require('plug-socket'));

/**
 * Flatten nested arrays.
 *
 *     flatten([ 4, 5, [ 8, 7 ] ])
 *     // → [ 4, 5, 8, 7 ]
 *
 * @param {Array} arrs
 */
function flatten (arrs) {
  return (ref = []).concat.apply(ref, arrs)
  var ref;
}

/**
 * Get the ID for an object.
 *
 * @param {object} item
 * @param {string} idProp
 */
function getId (item, idProp) {
  if ( idProp === void 0 ) idProp = 'id';

  return typeof item === 'object' ? item[idProp] : item
}

/**
 * Get IDs for an array of objects.
 *
 * @param {Array.<object>} items
 */
function getIds (items, idProp) {
  if ( idProp === void 0 ) idProp = 'id';

  const arr = Array.isArray(medias) ? medias : [ medias ];
  return arr.map((item) => getId(item, idProp))
}

/**
 * Parse a string timestamp from plug.dj.
 *
 * @param {string} timestamp
 */
function parseDate (timestamp) {
  return new Date(`${timestamp} UTC`)
}

/**
 * Fill in fn's arguments with ...args from the beginning of fn's arguments
 * list.
 *
 * @param {Function} fn
 * @param {...*} args
 */
function partial (fn) {
  var args = [], len = arguments.length - 1;
  while ( len-- > 0 ) args[ len ] = arguments[ len + 1 ];

  return function () {
  var rest = [], len = arguments.length;
  while ( len-- ) rest[ len ] = arguments[ len ];
 return fn.apply(void 0, args.concat( rest )) }
}

const ROLE = {
  NONE: 0,
  DJ: 1,
  BOUNCER: 2,
  MANAGER: 3,
  COHOST: 4,
  HOST: 5,

  AMBASSADOR: 3,
  ADMIN: 5
};

const MEDIA_SOURCE = {
  YOUTUBE: 1,
  SOUNDCLOUD: 2
};

const BAN_DURATION = {
  HOUR: 'h',
  DAY: 'd',
  PERMA: 'f'
};

const BAN_REASON = {
  SPAMMING: 1,
  VERBAL_ABUSE: 2,
  OFFENSIVE_PLAYS: 3,
  GENRE: 4,
  ATTITUDE: 5
};

const MUTE_DURATION = {
  SHORT: 's',
  MEDIUM: 'm',
  LONG: 'l'
};

const MUTE_REASON = {
  VIOLATING_RULES: 1,
  VERBAL_ABUSE: 2,
  SPAMMING: 3,
  LANGUAGE: 4,
  ATTITUDE: 5
};


var constants = Object.freeze({
	ROLE: ROLE,
	MEDIA_SOURCE: MEDIA_SOURCE,
	BAN_DURATION: BAN_DURATION,
	BAN_REASON: BAN_REASON,
	MUTE_DURATION: MUTE_DURATION,
	MUTE_REASON: MUTE_REASON
});

const debug$1 = createDebug('miniplug:backoff');

/**
 * Create a linear incremental backoff function. Returns a function that wraps
 * a Promise-returning function.
 *
 * **Usage**
 *
 *   const func = () => Promise.resolve('done')
 *   const backoff = createBackoff({ increment: 200, max: 2200 })
 *   const throttled = backoff(func)
 *   throttled().then(console.log) // Resolves immediately
 *   throttled().then(console.log) // Resolves after 200 ms
 */
function createBackoff (ref) {
  var increment = ref.increment;
  var max = ref.max;

  let queuedCalls = 0;
  let currentDelay = 0;
  let lastCall = Promise.resolve();

  function queueCall () {
    queuedCalls += 1;
    // Keep increasing the delay as more calls are queued.
    currentDelay = Math.min(currentDelay + increment, max);
    debug$1('queued', queuedCalls, currentDelay);
  }
  function unqueueCall () {
    queuedCalls -= 1;
    // If our queue is empty, reset the delay.
    if (queuedCalls === 0) {
      currentDelay = 0;
    }
    debug$1('unqueued', queuedCalls, currentDelay);
  }

  return (fn) => function () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    queueCall();

    debug$1('calling args', args.length);

    // `result` will resolve with the result of the function call.
    const result = lastCall.then(() => fn.apply(void 0, args));

    // `lastCall` will resolve after the backoff duration is over.
    lastCall = result
      .return(null) // Ignore return value of previous call.
      .catch(() => null) // Ignore errors.
      .delay(currentDelay)
      .tap(unqueueCall);

    return result
  }
}

/**
 * Reverse plug.dj's chat message escaping.
 */
function unescape (message) {
  return message
    .replace(/&#34;/g, '"')
    .replace(/&#39;/g, '\'')
    .replace(/&lt;/g, '<')
    .replace(/&gt;/g, '>')
    .replace(/&amp;/g, '&')
}

var unescape_1 = unescape;

/**
 * Add methods to an object's prototype.
 */
function makeProto (obj, methods) {
  const proto = Object.assign({}, {toJSON: () => obj},
    methods);
  const instance = Object.create(proto);
  Object.assign(instance, obj);
  return instance
}

function wrapMessage (mp, message) {
  if (message.un) {
    message.un = unescape_1(message.un);
  }

  message.id = message.cid;
  message.message = unescape_1(message.message);

  return makeProto(message, {
    own: () => mp.me().id === message.uid,

    chat: mp.chat,
    reply: partial(mp.chat, `@${message.un}`),
    emote: partial(mp.chat, `/me @${message.un}`),

    delete: partial(mp.deleteChat, message.cid),

    getUser () {
      const user = mp.user && mp.user(message.uid);
      return user ? Promise.resolve(user) : mp.getUser(message.uid)
    }
  })
}

function wrapHistoryEntry (mp, raw) {
  const entry = {
    id: raw.id,
    // wrapMedia expects a playlist ID, but we don't know it--pass null instead.
    media: mp.wrapMedia(null, raw.media),
    room: mp.wrapRoom(raw.room),
    user: mp.wrapUser(raw.user),
    time: parseDate(raw.timestamp),
    score: raw.score
  };

  return makeProto(entry, {
    getUser: partial(mp.getUser, raw.user.id)
  })
}

function wrapInventoryProduct (mp, product) {
  // Rename the product.id property. The `id` property on inventory products
  // corresponds to the `name` property on store products in the plug.dj API.
  // Miniplug uses `name` for both for consistency.
  product.name = product.id;
  delete product.id;

  return makeProto(product, {})
}

function wrapMedia (mp, playlistId, media) {
  return makeProto(media, {
    update: partial(mp.updateMedia, playlistId, media.id),
    delete: () =>
      mp.deleteMedia(playlistId, [ media.id ])
  })
}

function wrapNotification (mp, notif) {
  notif.id = Number(notif.id);
  notif.timestamp = parseDate(notif.timestamp);

  if (notif.action === 'levelUp') {
    notif.level = Number(notif.value);
  }

  if (notif.action === 'gift') {
    const parts = notif.value.split('\u2800');
    notif.from = parts[0];
    notif.amount = Number(parts[1]);
  }

  if (notif.action === 'custom') {
    notif.message = notif.value;
  }

  return makeProto(notif, {
    acknowledge: partial(mp.acknowledgeNotification, notif.id)
  })
}

function wrapPlaylist (mp, playlist) {
  return makeProto(playlist, {
    delete: partial(mp.deletePlaylist, playlist.id),
    activate: partial(mp.activatePlaylist, playlist.id),
    rename: partial(mp.renamePlaylist, playlist.id),
    shuffle: partial(mp.shufflePlaylist, playlist.id),

    getMedia: () =>
      mp.getMedia(playlist.id)
        .map(partial(mp.wrapMedia, playlist.id))
        .tap(media => {
          // cachedMedia = media
        }),
    insert: (media, append) =>
      {
        if ( append === void 0 ) append = true;

        return mp.insertMedia(playlist.id, media, append);
  },
    move: (media, before) =>
      mp.moveMedia(playlist.id, media, before)
  })
}

function wrapRoom (mp, room) {
  if (room.welcome) {
    room.welcome = unescape_1(room.welcome);
  }
  if (room.description) {
    room.description = unescape_1(room.description);
  }

  room.isFavorite = room.favorite;
  delete room.favorite;

  return makeProto(room, {
    getHost: partial(mp.getUser, room.hostID),

    join: partial(mp.join, room.slug),

    favorite: partial(mp.favoriteRoom, room.id),
    unfavorite: partial(mp.unfavoriteRoom, room.id)
  })
}

function wrapStoreProduct (mp, product) {
  return makeProto(product, {
    purchase: partial(mp.purchase, product.id)
  })
}

function wrapUser (mp, user) {
  user.username = user.guest ? null : unescape_1(user.username || '');

  // Chat mention string for this user.
  const mention = `@${user.username}`;

  return makeProto(user, {
    hasPermission: (role) =>
      (user.role >= role || user.gRole >= role),
    hasGlobalPermission: (role) =>
      user.gRole >= role,

    chat: partial(mp.chat, mention),
    emote: partial(mp.chat, `/me ${mention}`),

    add: partial(mp.addDJ, user.id),
    move: partial(mp.moveDJ, user.id),
    remove: partial(mp.removeDJ, user.id),
    skip: partial(mp.skipDJ, user.id),

    befriend: partial(mp.befriend, user.id),
    rejectRequest: partial(mp.rejectFriendRequest, user.id),

    ignore: partial(mp.ignore, user.id),
    unignore: partial(mp.unignore, user.id),

    mute: partial(mp.mute, user.id),
    unmute: partial(mp.unmute, user.id),

    ban: partial(mp.ban, user.id),
    unban: partial(mp.unban, user.id),

    getHistory: partial(mp.getUserHistory, user.id),

    mention: () => mention,
    // Allows mentioning users by doing `Hello ${user}` in chat messages.
    toString: () => mention
  })
}

// Have to use a real class to get working inheritance from Array. Methods are
// still added inside wrapWaitlist, as usual.
class Waitlist extends Array {}

function wrapWaitlist (mp, waitlist) {
  const wrapped = Object.assign(new Waitlist(), {
    contains: (user) =>
      wrapped.some((waiting) => waiting.id === getId(user)),
    positionOf: (user) =>
      wrapped.findIndex((waiting) => waiting.id === getId(user)),
    toJSON: () => waitlist
  });

  waitlist.forEach((id) => {
    wrapped.push(mp.user(id) || mp.wrapUser({ id: id }));
  });

  return wrapped
}

function dataModelPlugin () {
  return (mp) => {
    Object.assign(mp, {
      wrapMessage: partial(wrapMessage, mp),
      wrapHistoryEntry: partial(wrapHistoryEntry, mp),
      wrapInventoryProduct: partial(wrapInventoryProduct, mp),
      wrapMedia: partial(wrapMedia, mp),
      wrapNotification: partial(wrapNotification, mp),
      wrapPlaylist: partial(wrapPlaylist, mp),
      wrapRoom: partial(wrapRoom, mp),
      wrapStoreProduct: partial(wrapStoreProduct, mp),
      wrapUser: partial(wrapUser, mp),
      wrapWaitlist: partial(wrapWaitlist, mp)
    });
  }
}

const debug$2 = createDebug('miniplug:http');

function httpPlugin (httpOpts) {
  httpOpts = Object.assign({}, {backoff: (fn) => fn},
    httpOpts);

  return (mp) => {
    const request = httpOpts.backoff(
      // wait until connections are complete before sending off requests
      (url, opts) => mp.connected
        .tap(() => debug$2(opts.method, url, opts.body || opts.query))
        .then((session) =>
          got(`${httpOpts.host}/_/${url}`, Object.assign({}, {headers: {
              cookie: session.cookie,
              'content-type': 'application/json'
            },
            json: true},
            opts,
            {body: opts.body ? JSON.stringify(opts.body) : undefined}))
        )
        .then((resp) => {
          if (resp.body.status !== 'ok') {
            throw new Error(resp.body.data.length ? resp.body.data[0] : resp.body.status)
          }
          return resp.body.data
        })
    );

    const post = (url, data) => request(url, { method: 'post', body: data });
    const get = (url, data) => request(url, { method: 'get', query: data });
    const put = (url, data) => request(url, { method: 'put', body: data });
    const del = (url, data) => request(url, { method: 'delete', body: data });

    mp.request = request;
    mp.post = post;
    mp.get = get;
    mp.put = put;
    mp.del = del;
  }
}

const debug$3 = createDebug('miniplug:connect');
const debugWs = createDebug('miniplug:ws');

function connectPlugin (options) {
  if ( options === void 0 ) options = {};

  return (mp) => {
      // log in
    const loginOpts = { host: options.host, authToken: true };

    function connect (opts) {
      debug$3('connecting', opts.email);

      const loginPromise = Promise.resolve(
        opts.email
          ? login.user(opts.email, opts.password, loginOpts)
          : login.guest(loginOpts)
      );

      const ws = socket();
      ws.setMaxListeners(100);
      ws.on('action', (type, payload) => {
        debugWs(type, payload);
      });

      const connected = loginPromise
        .then((res) => new Promise((resolve, reject) => {
          ws.auth(res.token);
          ws.once('error', reject);

          mp.isConnected = true;
          mp.emit('login');

          const me = mp.getMe();
          ws.once('ack', () => {
            resolve({ cookie: res.cookie });
            ws.removeListener('error', reject);

            me.then((user) => mp.emit('connected', user));
          });
        }))
        .catch((err) => {
          mp.emit('error', err);
          throw err
        });

      mp.ws = ws;
      mp.connected = connected;

      return connected
    }

    mp.connect = connect;
  }
}

const debug$4 = createDebug('miniplug:users');

const GUEST_ID = 0;

function usersPlugin () {
  const currentGuestsCount = Symbol('Guests count');
  const currentUsers = Symbol('Users');
  const currentUser = Symbol('Me');

  return (mp) => {
    // local user cache/collection API
    mp[currentGuestsCount] = 0;
    mp[currentUsers] = [];

    function onUserJoin (user) {
      debug$4('join', user.id);
      if (user.guest) {
        mp[currentGuestsCount] += 1;
        mp.emit('guestJoin');
      } else {
        user = mp.wrapUser(user);
        mp[currentUsers].push(user);
        mp.emit('userJoin', user);
      }
    }

    function onUserLeave (id) {
      debug$4('leave', id);
      if (id === GUEST_ID) {
        if (mp[currentGuestsCount] > 0) {
          mp[currentGuestsCount] -= 1;
        }
        mp.emit('guestLeave');
      } else {
        const i = mp[currentUsers].findIndex((user) => user.id === id);
        if (i !== -1) {
          const user = mp[currentUsers][i];
          mp[currentUsers].splice(i, 1);
          mp.emit('userLeave', user);
        }
      }
    }

    function onUserUpdate (props) {
      const user = mp.user(props.i);
      if (!user) { return }

      delete props.i;
      const oldProps = Object.keys(props).reduce((o, name) => {
        o[name] = user[name];
        return o
      }, {});

      Object.assign(user, props);

      mp.emit('userUpdate', user, oldProps);
    }

    // Emit `userUpdate` events for `role` updates.
    function onStaffUpdate (props) {
      props.u.forEach((ref) => {
        var i = ref.i;
        var p = ref.p;

        onUserUpdate({ i, role: p });
      });
    }

    mp.on('connected', (user) => {
      mp[currentUser] = mp.wrapUser(user);

      mp.ws.on('userJoin', onUserJoin);
      mp.ws.on('userLeave', onUserLeave);
      mp.ws.on('userUpdate', onUserUpdate);
      mp.ws.on('modStaff', onStaffUpdate);
    });

    // keeping things in sync
    mp.on('roomState', (ref) => {
      var users = ref.users;
      var guests = ref.meta.guests;

      mp[currentUsers] = users.map(mp.wrapUser);
      mp[currentGuestsCount] = guests;
    });

    const me = () => mp[currentUser];
    const user = (id) => {
      if (mp[currentUser] && id === mp[currentUser].id) {
        return mp[currentUser]
      }
      return mp[currentUsers].find((user) => user.id === id)
    };
    // TODO May want to include `me()` in the `users()` list in v2.0.0. I'm not
    // sure which is more expected.
    const users = () => mp[currentUsers];
    const guests = () => mp[currentGuestsCount];

    // REST API
    const getMe = () => mp.get('users/me').get(0).then(mp.wrapUser);
    const getUser = uid => mp.get(`users/${uid}`).get(0).then(mp.wrapUser);
    // pass IDs or arrays of IDs
    const getUsers = () =>
      {
        var uids = [], len = arguments.length;
        while ( len-- ) uids[ len ] = arguments[ len ];

        return mp.post('users/bulk', {
        ids: flatten(uids)
      }).map(mp.wrapUser);
    };

    // current user & profile
    const saveSettings = partial(mp.put, 'users/settings');
    const setAvatar = (avatar) =>
      mp.put('users/avatar', { id: avatar });
    const setBadge = (badge) =>
      mp.put('users/badge', { id: badge });
    const setBlurb = (blurb) =>
      mp.put('profile/blurb', { blurb: blurb });
    const setLanguage = (lang) =>
      mp.put('users/language', { language: lang });
    const getTransactions = partial(mp.get, 'users/me/transactions');

    const validateUsername = (name) =>
      mp.get(`users/validate/${encodeURIComponent(name)}`)
        .get(0)
        .catch(() => Promise.reject(new Error('Username unavailable.')));

    // Public API
    Object.assign(mp, {
      // local
      me,
      user,
      users,
      guests,
      // remote
      getMe,
      getUser,
      getUsers,
      // setters
      saveSettings,
      setAvatar,
      setBadge,
      setBlurb,
      setLanguage,

      getTransactions,
      validateUsername
    });
  }
}

function ignoresPlugin () {
  return (mp) => {
    const getIgnoredUsers = partial(mp.get, 'ignores');
    const ignore = (uid) =>
      mp.post('ignores', { id: uid }).get(0);
    const unignore = (uid) =>
      mp.del(`ignores/${uid}`);

    Object.assign(mp, {
      getIgnoredUsers,
      ignore,
      unignore
    });
  }
}

function mutesPlugin () {
  return (mp) => {
    const getMutes = partial(mp.get, 'mutes');
    const mute = (uid, duration, reason) =>
      {
        if ( duration === void 0 ) duration = MUTE_DURATION.SHORT;
        if ( reason === void 0 ) reason = MUTE_REASON.VIOLATING_RULES;

        return mp.post('mutes', {
        userID: uid,
        duration: duration,
        reason: reason
      });
    };
    const unmute = (uid) =>
      mp.del(`mutes/${uid}`);

    Object.assign(mp, {
      getMutes,
      mute,
      unmute
    });
  }
}

function bansPlugin () {
  return (mp) => {
    const getBans = partial(mp.get, 'bans');
    const ban = (uid, duration, reason) =>
      {
        if ( duration === void 0 ) duration = BAN_DURATION.HOUR;
        if ( reason === void 0 ) reason = BAN_REASON.SPAMMING;

        return mp.post('bans/add', {
        userID: uid,
        reason: reason,
        duration: duration
      }).get(0);
    };
    const unban = (uid) =>
      mp.del(`bans/${uid}`);

    Object.assign(mp, {
      getBans,
      ban,
      unban
    });
  }
}

function notificationsPlugin () {
  const currentNotifications = Symbol('Notifications');

  return (mp) => {
    mp[currentNotifications] = [];

    function onNotify (notif) {
      mp[currentNotifications].push(notif);
      mp.emit('notify', mp.wrapNotification(notif));
    }

    mp.on('connected', (user) => {
      mp[currentNotifications] = (user && user.notifications) || [];

      mp.ws.on('notify', onNotify);
    });

    Object.assign(mp, {
      notifications: () => mp[currentNotifications].map(mp.wrapNotification),

      getNotifications: () =>
        mp.getMe()
          .then((me) => me.notifications || [])
          .tap((notifs) => {
            mp[currentNotifications] = notifs;
          })
          .map(mp.wrapNotification),

      acknowledgeNotification: (id) =>
        mp.del(`notifications/${id}`).tap(() => {
          // Remove the notification from the local notifications list.
          mp[currentNotifications] = mp[currentNotifications]
            .filter((notif) => notif.id !== Number(id));
        })
    });
  }
}

function boothPlugin (opts) {
  if ( opts === void 0 ) opts = {};

  const currentHistoryEntry = Symbol('History entry');

  return (mp) => {
    // Local state API
    const historyEntry = () => mp[currentHistoryEntry];
    const dj = () => mp[currentHistoryEntry] ? mp[currentHistoryEntry].dj : null;
    const media = () => mp[currentHistoryEntry] ? mp[currentHistoryEntry].media : null;

    mp.on('roomState', (state) => {
      mp[currentHistoryEntry] = {
        id: state.playback.historyID,
        dj: mp.user(state.booth.currentDJ),
        media: state.playback.media,
        playlistId: state.playback.playlistID,
        time: parseDate(state.playback.startTime)
      };
    });

    // Socket API
    function onAdvance (event) {
      const previous = historyEntry();
      if (!event || !event.m) {
        mp[currentHistoryEntry] = null;
        mp.emit('advance', null, previous);
        return
      }

      var historyId = event.h;
      var media = event.m;
      var djId = event.c;
      var playlistId = event.p;
      var time = event.t;

      time = parseDate(time);

      mp[currentHistoryEntry] = {
        id: historyId,
        dj: mp.user(djId),
        media: media,
        playlistId: playlistId,
        time: time
      };

      mp.emit('advance', historyEntry(), previous);
    }

    mp.on('connected', () => {
      mp.ws.on('advance', onAdvance);
    });

    // Rest API
    Object.assign(mp, {
      skipDJ: (uid, hid) =>
        mp.post('booth/skip', { userID: uid, historyID: hid }),
      skipMe: partial(mp.post, 'booth/skip/me')
    });

    Object.assign(mp, {
      historyEntry,
      dj,
      media
    });
  }
}

const debug$5 = createDebug('miniplug:waitlist');

function waitlistPlugin () {
  const currentWaitlist = Symbol('Waitlist');

  return (mp) => {
    mp[currentWaitlist] = mp.wrapWaitlist([]);

    mp.on('roomState', (state) => {
      mp[currentWaitlist] = mp.wrapWaitlist(state.booth.waitingDJs);
      mp.emit('waitlistUpdate', mp.waitlist(), []);
    });

    function onDjListUpdate (ids) {
      debug$5('update', ids);

      const previous = mp.waitlist();

      mp[currentWaitlist] = mp.wrapWaitlist(ids);
      mp.emit('waitlistUpdate', mp.waitlist(), previous);
    }

    function onAdvance (event) {
      if (event && event.d) {
        onDjListUpdate(event.d);
      } else {
        onDjListUpdate([]);
      }
    }

    function onModAddDj () {
      // TODO
    }
    function onModMoveDj () {
      // TODO
    }
    function onModRemoveDj () {
      // TODO
    }

    function onDjListCycle (ref) {
      var f = ref.f;
      var mi = ref.mi;

      mp.emit('waitlistCycle', {
        shouldCycle: f,
        user: mp.user(mi)
      });
    }

    function onDjListLocked (ref) {
      var f = ref.f;
      var c = ref.c;
      var mi = ref.mi;

      const user = mp.user(mi);
      mp.emit('waitlistLock', {
        locked: f,
        cleared: !!c,
        user
      });

      if (c) {
        mp.emit('waitlistClear', { user });
      }
    }

    mp.on('connected', () => {
      mp.ws.on('djListUpdate', onDjListUpdate);
      mp.ws.on('advance', onAdvance);
      mp.ws.on('modAddDJ', onModAddDj);
      mp.ws.on('modMoveDJ', onModMoveDj);
      mp.ws.on('modRemoveDJ', onModRemoveDj);
      mp.ws.on('djListCycle', onDjListCycle);
      mp.ws.on('djListLocked', onDjListLocked);
    });

    const setCycle = (val) =>
      {
        if ( val === void 0 ) val = true;

        return mp.put('booth/cycle', { shouldCycle: val });
    };

    Object.assign(mp, {
      waitlist: () => mp[currentWaitlist],

      joinWaitlist: partial(mp.post, 'booth'),
      leaveWaitlist: partial(mp.del, 'booth'),

      setCycle,
      enableCycle: partial(setCycle, true),
      disableCycle: partial(setCycle, false),

      setLock: (locked, clear) =>
        {
          if ( locked === void 0 ) locked = true;
          if ( clear === void 0 ) clear = false;

          return mp.put('booth/lock', { isLocked: locked, removeAllDJs: clear });
    },
      lockWaitlist: (clear) =>
        {
          if ( clear === void 0 ) clear = false;

          return mp.setLock(true, clear);
    },
      unlockWaitlist: () =>
        mp.setLock(false, false),

      addDJ: (uid) =>
        mp.post('booth/add', { id: uid }),
      moveDJ: (uid, pos) =>
        mp.post('booth/move', { userID: uid, position: pos }),
      removeDJ: (uid) =>
        mp.del(`booth/remove/${uid}`)
    });
  }
}

function historyPlugin () {
  return (mp) => {
    const getRoomHistory = () =>
      mp.get('rooms/history').map(mp.wrapHistoryEntry);

    // Get the history for the logged-in user.
    const getOwnHistory = () =>
      mp.get('users/me/history').map(mp.wrapHistoryEntry);

    const getProfile = (id) =>
      mp.get(`profile/${id}`).get(0);

    // Get the history for another user.
    const getOtherHistory = (id) =>
      getProfile(id).get('history').map(mp.wrapHistoryEntry);

    // Get the history for a user. Defaults to the current user.
    const getUserHistory = (id) =>
      {
        if ( id === void 0 ) id = mp.me().id;

        return mp.me().id === id ? getOwnHistory() : getOtherHistory(id);
    };

    Object.assign(mp, {
      getRoomHistory,
      getUserHistory
    });
  }
}

const debug$6 = createDebug('miniplug:chat');

function chatPlugin (opts) {
  opts = Object.assign({}, {timeout: 7000,
    // allow users to pass their own rate limiting function,
    // this will get a sensible default at some point
    backoff: (fn) => fn},
    opts);

  return (mp) => {
    // translate raw socket events to wrapped miniplug events
    function onChat (msg) {
      debug$6('chat', msg.uid, msg.un, msg.message);
      mp.emit('chat', mp.wrapMessage(msg));
    }
    function onChatDelete (ref) {
      var c = ref.c;
      var mi = ref.mi;

      debug$6('chatDelete', mi, c);
      mp.emit('chatDelete', {
        cid: c,
        user: mp.user(mi)
      });
    }

    mp.on('connected', () => {
      mp.ws.on('chat', onChat);
      mp.ws.on('chatDelete', onChatDelete);
    });

    // REST API
    const deleteChat = (cid) =>
      mp.del(`chat/${cid}`);

    // Socket API
    const chat = opts.backoff(function () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      const ws = mp.ws;
      const message = args.join(' ');
      debug$6('send', message);
      ws.chat(message);
      // attempt to resolve when the message comes back
      return new Promise((resolve, reject) => {
        const intv = setTimeout(() => {
          mp.removeListener('chat', onChat);
          reject(new Error('Waiting for chat message to come back timed out.'));
        }, opts.timeout);

        const onChat = (chat) => {
          debug$6('is same?', chat.own(), message, chat.message);
          // TODO ensure that chat messages are unescaped enough for this
          // (they aren't)
          if (chat.own() && chat.message === message) {
            mp.removeListener('chat', onChat);
            clearTimeout(intv);
            resolve(chat);
          }
        };

        mp.on('chat', onChat);
      })
    });
    const emote = partial(chat, '/me');

    // Public API
    Object.assign(mp, {
      deleteChat,
      chat,
      emote
    });
  }
}

function friendsPlugin () {
  return (mp) => {
    // Friendship events only include the requesting/accepting user's name.
    // Plug.dj refreshes its client-side friends or invites list when these
    // events come in. Miniplug requests the friends or invites list and emits
    // the events with a user object, so it's easy to respond to the events
    // using the `user.befriend` and `user.rejectRequest` methods.
    function onFriendRequest (name) {
      getFriendRequests().each((request) => {
        if (unescape_1(request.username) === unescape_1(name)) {
          mp.emit('friendRequest', mp.wrapUser(request));
        }
      });
    }

    function onFriendAccept (name) {
      getFriends().each((friend) => {
        if (friend.username === unescape_1(name)) {
          mp.emit('friendAccept', friend);
        }
      });
    }

    mp.on('connected', () => {
      mp.ws.on('friendRequest', onFriendRequest);
      mp.ws.on('friendAccept', onFriendAccept);
    });

    // REST Friend API
    const getFriends = () =>
      mp.get('friends').map(mp.wrapUser);
    const befriend = (uid) =>
      mp.post('friends', { id: uid });
    const unfriend = (uid) =>
      mp.del(`friends/${uid}`);
    const getFriendRequests = partial(mp.get, 'friends/invites');
    const rejectFriendRequest = (uid) =>
      mp.put('friends/ignore', { id: uid });

    // Public API
    Object.assign(mp, {
      getFriends,
      befriend,
      unfriend,
      getFriendRequests,
      rejectFriendRequest
    });
  }
}

const debug$7 = createDebug('miniplug:rooms');

function roomsPlugin () {
  const currentRoom = Symbol('Current room');

  return (mp) => {
    mp[currentRoom] = null;

    mp.on('roomState', (state) => {
      mp[currentRoom] = mp.wrapRoom(state.meta);
    });

    // Add a handler to process a room property update.
    //
    //  * `eventName` - Plug.dj socket event name for the property.
    //  * `sockProp` - Property name of the new value on the plug.dj socket
    //        event parameter.
    //  * `roomProp` - Property name for the value on the miniplug room object.
    function addUpdateHandler (eventName, sockProp, roomProp) {
      mp.ws.on(eventName, (data, targetSlug) => {
        const user = mp.user(data.u);
        const value = data[sockProp];

        debug$7(eventName, user && user.id, value);

        if (mp[currentRoom] && mp[currentRoom].slug === targetSlug) {
          mp[currentRoom][roomProp] = value;
        }

        mp.emit(eventName, value, user);
        mp.emit('roomUpdate', {
          [roomProp]: value
        }, user);
      });
    }

    mp.on('connected', () => {
      addUpdateHandler('roomNameUpdate', 'n', 'name');
      addUpdateHandler('roomDescriptionUpdate', 'd', 'description');
      addUpdateHandler('roomWelcomeUpdate', 'w', 'welcome');
      addUpdateHandler('roomMinChatLevelUpdate', 'm', 'minChatLevel');
    });

    const room = () => mp[currentRoom];

    const getRooms = (query, page, limit) =>
      {
        if ( query === void 0 ) query = '';
        if ( page === void 0 ) page = 0;
        if ( limit === void 0 ) limit = 50;

        return mp.get('rooms', { q: query, page, limit })
        .map(mp.wrapRoom);
    };
    const getFavorites = (query, page, limit) =>
      {
        if ( query === void 0 ) query = '';
        if ( page === void 0 ) page = 0;
        if ( limit === void 0 ) limit = 50;

        return mp.get('rooms/favorites', { q: query, page, limit })
        .map(mp.wrapRoom);
    };
    const createRoom = (name, isPrivate) =>
      {
        if ( isPrivate === void 0 ) isPrivate = false;

        return mp.post('rooms', { name: name, private: isPrivate }).get(0);
    };
    const getMyRooms = () =>
      mp.get('rooms/me').map(mp.wrapRoom);

    const favoriteRoom = (rid) =>
      mp.post('rooms/favorites', { id: rid });
    const unfavoriteRoom = (rid) =>
      mp.del(`rooms/favorites/${rid}`);
    const join = (slug) =>
      mp.post('rooms/join', { slug: slug }).then(getRoomState);
    const getRoomState = () =>
      mp.get('rooms/state').get(0)
        .tap(mp.emit.bind(mp, 'roomState'));

    const updateRoom = (patch) => {
      if (!room()) {
        return Promise.reject(new Error('You are not currently in a room.'))
      }
      // TODO check keys used in `patch`? Only 'name', 'description', 'welcome'
      // and 'minChatLevel' can be updated.
      return mp.post('rooms/update', patch)
    };

    const setRoomWelcome = (welcome) => updateRoom({ welcome });
    const setRoomDescription = (description) => updateRoom({ description });
    const setRoomName = (name) => updateRoom({ name });
    const setRoomMinChatLevel = (minChatLevel) => updateRoom({ minChatLevel });

    const validateRoomName = (name) =>
      mp.get(`rooms/validate/${encodeURIComponent(name)}`)
        .get(0)
        .catch(() => Promise.reject(new Error('Room name unavailable.')));

    Object.assign(mp, {
      // local
      room,
      // remote
      getRooms,
      getFavorites,
      getMyRooms,
      createRoom,
      validateRoomName,
      setRoomWelcome,
      setRoomDescription,
      setRoomName,
      setRoomMinChatLevel,
      // favorites
      favoriteRoom,
      unfavoriteRoom,
      // joining
      join,
      getRoomState
    });
  }
}

function playlistsPlugin () {
  return (mp) => {
    mp.on('connected', () => {
      mp.ws.on('playlistCycle', (id) => {
        mp.emit('playlistCycle', id);
      });
    });

    Object.assign(mp, {
      getPlaylists: () =>
        mp.get('playlists').map(mp.wrapPlaylist),
      getActivePlaylist: () =>
        mp.get('playlists')
          .filter((playlist) => playlist.active)
          .get(0).then(mp.wrapPlaylist),
      createPlaylist: (name/*, initialMedia */) =>
        mp.post('playlists', { name: name }),
      deletePlaylist: (pid) =>
        mp.del(`playlists/${pid}`),
      activatePlaylist: (pid) =>
        mp.put(`playlists/${pid}/activate`),
      renamePlaylist: (pid, name) =>
        mp.put(`playlists/${pid}/rename`, { name: name }),
      shufflePlaylist: (pid) =>
        mp.put(`playlists/${pid}/shuffle`),

      getMedia: (pid) =>
        mp.get(`playlists/${pid}/media`).map(partial(mp.wrapMedia, pid)),
      updateMedia: (pid, mid, author, title) =>
        mp.put(`playlists/${pid}/media/update`, {
          id: mid,
          author: author,
          title: title
        }).get(0),
      moveMedia: (pid, mids, before) =>
        mp.put(`playlists/${pid}/media/move`, {
          ids: getIds(mids),
          beforeID: getId(before)
        }),
      insertMedia: (pid, media, append) =>
        {
          if ( append === void 0 ) append = true;

          return mp.post(`playlists/${pid}/media/insert`, {
          media: media,
          append: append
        });
    },
      deleteMedia: (pid, mids) =>
        mp.post(`playlists/${pid}/media/delete`, {
          ids: getIds(mids)
        })
    });
  }
}

function storePlugin () {
  return (mp) => {
    const getProducts = (type, category) =>
      {
        if ( category === void 0 ) category = 'all';

        return mp.get(`store/products/${type}/${category}`).map(mp.wrapStoreProduct);
    };
    const getStoreAvatars = partial(getProducts, 'avatars');
    const getStoreBadges = partial(getProducts, 'badges');
    const getStoreMisc = partial(getProducts, 'misc');

    const getInventory = (type) =>
      mp.get(`store/inventory/${type}`).map(mp.wrapInventoryProduct);
    const getOwnedAvatars = partial(getInventory, 'avatars');
    const getOwnedBadges = partial(getInventory, 'badges');

    const purchase = (product) =>
      mp.post('store/purchase', { id: getId(product) }).get(0);

    const getNameChangeProductId = () =>
      getStoreMisc('username').get(0).get('id');
    const purchaseNameChange = (username) =>
      Promise.props({
        id: getNameChangeProductId(),
        username
      }).then(
        partial(mp.post, 'store/purchase/username')
      );

    Object.assign(mp, {
      getProducts,
      getStoreAvatars,
      getStoreBadges,
      getStoreMisc,
      getInventory,
      getOwnedAvatars,
      getOwnedBadges,
      purchase,
      purchaseNameChange
    });
  }
}

function systemPlugin () {
  return (mp) => {
    function onMessage (text) {
      mp.emit('systemMessage', text);
    }

    function onMaintenanceAlert (minutesLeft) {
      mp.emit('maintenanceAlert', minutesLeft);
    }

    function onMaintenance () {
      mp.emit('maintenance');
    }

    function onUpdate () {
      mp.emit('systemUpdate');
    }

    mp.on('connected', () => {
      mp.ws.on('plugMessage', onMessage);
      mp.ws.on('plugMaintenanceAlert', onMaintenanceAlert);
      mp.ws.on('plugMaintenance', onMaintenance);
      mp.ws.on('plugUpdate', onUpdate);
    });
  }
}

function votePlugin (opts) {
  if ( opts === void 0 ) opts = {};

  const currentVoteStats = Symbol('Votes');
  const currentGrabs = Symbol('Grabs');

  return function (mp) {
    mp[currentVoteStats] = [];
    mp[currentGrabs] = [];

    mp.on('advance', () => {
      mp[currentVoteStats] = [];
      mp[currentGrabs] = [];
    });

    function onVote (ref) {
      var i = ref.i;
      var v = ref.v;

      mp[currentVoteStats].push({
        uid: i,
        vote: v
      });
      const user = mp.user(i) || mp.wrapUser({ id: i });
      mp.emit('vote', {
        user,
        vote: v
      });
    }

    function onGrab (uid) {
      mp[currentGrabs].push(uid);
      const user = mp.user(uid);
      mp.emit('grab', user);
    }

    mp.on('connected', () => {
      mp.ws.on('vote', onVote);
      mp.ws.on('grab', onGrab);
    });

    mp.on('roomState', (state) => {
      mp[currentGrabs] = Object.keys(state.grabs);
      mp[currentVoteStats] = Object.keys(state.votes).map((uid) => ({
        uid,
        vote: state.votes[uid]
      }));
    });

    mp.score = () => {
      const map = mp[currentVoteStats].reduce(
        (map, v) => Object.assign(map, { [v.uid]: v.vote }),
        {}
      );
      let positive = 0;
      let negative = 0;
      Object.keys(map).forEach((uid) => {
        if (map[uid] === 1) {
          positive++;
        } else if (map[uid] === -1) {
          negative++;
        }
      });
      return {
        positive,
        negative,
        grabs: mp[currentGrabs].length,
        listeners: mp.users().length + mp.guests()
      }
    };

    const vote = (direction) =>
      mp.post('votes', {
        historyID: mp.historyEntry().id,
        direction: direction
      });

    Object.assign(mp, {
      grab: (targetPlaylist, hid) =>
        mp.post('grabs', { playlistID: targetPlaylist, historyID: hid }).get(0),

      vote,

      woot: partial(vote, 1),
      meh: partial(vote, -1)
    });
  }
}

// Exports

Object.assign(miniplug$1, Object.assign({}, {httpPlugin,
  connectPlugin,
  usersPlugin,
  ignoresPlugin,
  mutesPlugin,
  bansPlugin,
  notificationsPlugin,
  boothPlugin,
  waitlistPlugin,
  historyPlugin,
  chatPlugin,
  friendsPlugin,
  roomsPlugin,
  playlistsPlugin,
  storePlugin,
  systemPlugin,
  votePlugin},
  constants));

// Implementation

const debug = createDebug('miniplug:miniplug');
const defaultOptions = {
  host: 'https://plug.dj',
  connect: true
};

function miniplug$1 (opts) {
  if ( opts === void 0 ) opts = {};

  // Faux-inherit from EventEmitter.
  const emitter = new events.EventEmitter();
  emitter.setMaxListeners(100);
  const mp = Object.create(emitter);

  opts = Object.assign({}, defaultOptions, opts);

  // trim trailing slashes
  const plugHost = opts.host.replace(/\/+$/, '');

  mp.on('login', () => debug('authenticated'));
  mp.on('connected', () => debug('connected'));

  // Super-Duper Advanced Plugin API
  function use (plugin) {
    plugin(mp);
    return mp
  }

  // make miniplug!
  mp.use = use;

  use(dataModelPlugin());
  use(httpPlugin({
    host: plugHost,
    // This is the same backoff as used in Sooyou/plugged:
    // https://github.com/SooYou/plugged/blob/856bd0ef47307491c0ad95cba7006cd4721828fd/query.js#L4
    // And that seems pretty robust.
    backoff: createBackoff({ increment: 200, max: 2200 })
  }));
  use(connectPlugin({
    host: plugHost
  }));
  use(usersPlugin());
  use(ignoresPlugin());
  use(mutesPlugin());
  use(bansPlugin());
  use(notificationsPlugin());
  use(boothPlugin());
  use(waitlistPlugin());
  use(historyPlugin());
  use(chatPlugin({
    backoff: createBackoff({ increment: 70, max: 700 })
  }));
  use(friendsPlugin());
  use(roomsPlugin());
  use(playlistsPlugin());
  use(storePlugin());
  use(systemPlugin());
  use(votePlugin());

  // Misc
  Object.assign(mp, {
    // REST: News APIs
    getNews: partial(mp.get, 'news'),
  });

  if (opts.connect) { mp.connect(opts); }

  return mp
}

module.exports = miniplug$1;
//# sourceMappingURL=index.js.map
